#ifndef CODE_sjdbPrepare
#define CODE_sjdbPrepare

#include "SjdbClass.h"
#include "Parameters.h"
#include "Genome.h"

void sjdbPrepare (SjdbClass &sjdbLoci, Parameters &P, uint nGenomeReal, string outDir, Genome &mapGen, char *Gsj);

#endif