#ifndef CODE_genomeParametersWrite
#define CODE_genomeParametersWrite

#include "Parameters.h"
#include "Genome.h"

void genomeParametersWrite(string fileName, Parameters& P, string errorOut, Genome &mapGen);

#endif