#ifndef H_twoPassRunPass1
#define H_twoPassRunPass1

#include "Parameters.h"
#include "Genome.h"
#include "Transcriptome.h"
#include "SjdbClass.h"

void twoPassRunPass1(Parameters &P, Genome &genomeMain, Transcriptome *transcriptomeMain, SjdbClass &sjdbLoci);

#endif