#ifndef H_bamSortByCoordinate
#define H_bamSortByCoordinate

#include "Parameters.h"
#include "ReadAlignChunk.h"
#include "Genome.h"
#include "Solo.h"

void bamSortByCoordinate(Parameters &P, ReadAlignChunk **RAchunk, Genome &genome, Solo &solo) ;

#endif
