#ifndef CODE_sjdbBuildIndex
#define CODE_sjdbBuildIndex

#include "Parameters.h"
#include "PackedArray.h"
#include "Genome.h"

void sjdbBuildIndex (Parameters &P, char *Gsj, char *G, PackedArray &SA, PackedArray &SA2, PackedArray &SAi, Genome &mapGen, Genome &mapGen1);

#endif
